import copy
import datetime
import re

from .base import BaseModel
import sys

if sys.version_info < (3, 9):
    import typing

_re_date_format = re.compile(r'^\d\d\d\d-\d\d-\d\d$')
_re_datetime_format = re.compile(
    r'^(\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d)\+(\d\d):(\d\d)$')


def _datetime_value(values: dict, key: str) -> datetime.datetime or None:
    if key in values and values[key] is not None:
        value = values[key]
        match = _re_datetime_format.fullmatch(value)
        if match is not None:
            (dt, tz_hours, tz_minutes) = match.groups()
            value = "{}+{}{}".format(dt, tz_hours, tz_minutes)
            return datetime.datetime.strptime(
                value, '%Y-%m-%dT%H:%M:%S%z')

    return None


def _date_value(values: dict, key: str) -> datetime.date or None:
    if key in values and values[key] is not None:
        if _re_date_format.match(values[key]) is not None:
            return datetime.datetime.strptime(
                values[key], '%Y-%m-%d').date()

    return None


def _string_value(values: dict, key: str) -> str:
    if key in values and values[key]:
        return str(values[key])
    return ''


def _int_value(values: dict, key: str) -> int:
    if key in values and values[key]:
        return int(values[key])
    return 0


def _list_value(values: dict, key: str) -> list:
    if key in values and type(values[key]) is list:
        return copy.deepcopy(values[key])
    return []


def _timestamp2datetime(timestamp: int) -> datetime.datetime or None:
    if timestamp is not None:
        return datetime.datetime.fromtimestamp(timestamp)
    return None


class Response(BaseModel):
    domains_count: int
    if sys.version_info < (3, 9):
        domains_list: typing.List[str]
    else:
        domains_list: [str]

    def __init__(self, values):
        super().__init__()

        self.domains_count = 0
        self.domains_list = []

        if values is not None:
            self.domains_count = _int_value(values, 'domainsCount')
            self.domains_list = _list_value(values, 'domainsList')


class ErrorMessage(BaseModel):
    code: int
    message: str

    def __init__(self, values):
        super().__init__()

        self.int = 0
        self.message = ''

        if values is not None:
            self.code = _int_value(values, 'code')
            self.message = _string_value(values, 'messages')



from requests import request, Response
from ..exceptions.error import ApiAuthError, HttpApiError, BadRequestError
from ..version import VERSION, LIBRARY_NAME
import logging


class ApiRequester:
    __logger = logging.getLogger("api-requester")
    __connect_timeout = 10
    __user_agent = "{name}/{ver}".format(name=LIBRARY_NAME, ver=VERSION)
    _base_url: str
    _timeout: float

    def __init__(self, **kwargs):
        """

        :param kwargs: Supported parameters:
        - base_url: (optional) API endpoint URL; str
        - timeout: (optional) API call timeout in seconds; float
        """
        self._base_url = ''
        self.timeout = 30

        if 'base_url' in kwargs:
            self.base_url = kwargs['base_url']
        if 'timeout' in kwargs:
            self.timeout = kwargs['timeout']

    @property
    def base_url(self) -> str:
        return self._base_url

    @base_url.setter
    def base_url(self, url: str):
        if url is None or len(url) <= 8 or not url.startswith('http'):
            raise ValueError("Invalid URL specified.")
        self._base_url = url

    @property
    def timeout(self) -> float:
        """API call timeout in seconds"""
        return self._timeout

    @timeout.setter
    def timeout(self, value: float):
        """API call timeout in seconds"""
        if value is not None and 1 <= value <= 60:
            self._timeout = value
        else:
            raise ValueError("Timeout value should be in [1, 60]")

    def get(self, payload: dict) -> str:
        headers = {
            'User-Agent': ApiRequester.__user_agent,
            'Connection': 'close'
        }
        response = request(
            "GET",
            self.base_url,
            params=payload,
            headers=headers,
            timeout=(ApiRequester.__connect_timeout, self.timeout)
        )

        return ApiRequester._handle_response(response)

    def post(self, data: dict) -> str:
        headers = {
            'User-Agent': ApiRequester.__user_agent,
            'Connection': 'close'
        }

        response = request(
            'POST',
            self.base_url,
            json=data,
            headers=headers,
            timeout=(ApiRequester.__connect_timeout, self.timeout)
        )

        return ApiRequester._handle_response(response)

    @staticmethod
    def _handle_response(response: Response) -> str:
        if 200 <= response.status_code < 300:
            return response.content.decode('UTF-8')

        if response.status_code in [401, 402, 403]:
            raise ApiAuthError(response.text)

        if response.status_code in [400, 422]:
            raise BadRequestError(response.text)

        if response.status_code >= 300:
            raise HttpApiError(response.text)



__all__ = ['ApiAuthError', 'ApiRequester', 'BadRequestError', 'Client',
           'DomainDiscoveryApiError', 'EmptyApiKeyError', 'ErrorMessage',
           'HttpApiError', 'ParameterError', 'Response', 'ResponseError',
           'UnparsableApiResponseError']

from .client import Client
from .net.http import ApiRequester
from .models.response import ErrorMessage, Response

from .exceptions.error import DomainDiscoveryApiError, ParameterError, \
    EmptyApiKeyError, ResponseError, UnparsableApiResponseError, \
    ApiAuthError, BadRequestError, HttpApiError



import unittest
from json import loads
from domaindiscovery import Response, ErrorMessage


_json_response_ok_empty = '''{
   "domainsCount": 0,
   "domainsList": []
}'''

_json_response_ok = '''{
   "domainsCount": 2,
   "domainsList": [
        "aws.amazon.fr",
        "aws.amazon.org"
    ]
}'''

_json_response_error = '''{
    "code": 403,
    "messages": "Access restricted. Check credits balance or enter the correct API key."
}'''


class TestModel(unittest.TestCase):

    def test_response_parsing(self):
        response = loads(_json_response_ok)
        parsed = Response(response)
        self.assertEqual(parsed.domains_count, response['domainsCount'])
        self.assertIsInstance(parsed.domains_list, list)

        self.assertEqual(parsed.domains_list[0], response['domainsList'][0])

    def test_error_parsing(self):
        error = loads(_json_response_error)
        parsed_error = ErrorMessage(error)
        self.assertEqual(parsed_error.code, error['code'])
        self.assertEqual(parsed_error.message, error['messages'])



class BaseModel:
    def __init__(self):
        pass

    def __str__(self):
        result = {}
        for k, v in self.__dict__.items():
            result[k] = str(v)
        return str(result)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        is_equal = isinstance(other, self.__class__)
        for k, v in self.__dict__.values():
            is_equal = is_equal and (other.__dict__.get(k) == v)

        return is_equal

    def __getitem__(self, item):
        if type(item) is str and item in self.__dict__:
            return self.__dict__[item]
        raise KeyError("Invalid key: {}".format(item))



from json import loads
from ..models.response import ErrorMessage


class DomainDiscoveryApiError(Exception):
    def __init__(self, message):
        self.message = message

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, message):
        self._message = message

    def __str__(self):
        return str(self.__dict__)


class ParameterError(DomainDiscoveryApiError):
    pass


class EmptyApiKeyError(DomainDiscoveryApiError):
    pass


class ResponseError(DomainDiscoveryApiError):
    def __init__(self, message):
        self.message = message
        self._parsed_message = None
        try:
            parsed = loads(message)
            self.parsed_message = ErrorMessage(parsed)
        except Exception:
            pass

    @property
    def parsed_message(self):
        return self._parsed_message

    @parsed_message.setter
    def parsed_message(self, pm):
        self._parsed_message = pm


class UnparsableApiResponseError(DomainDiscoveryApiError):
    def __init__(self, message, origin_error):
        self.message = message
        self.original_error = origin_error

    @property
    def original_error(self):
        return self._original_error

    @original_error.setter
    def original_error(self, oe):
        self._original_error = oe


class ApiAuthError(ResponseError):
    pass


class BadRequestError(ResponseError):
    pass


class HttpApiError(DomainDiscoveryApiError):
    pass



__all__ = ['ParameterError', 'HttpApiError', 'DomainDiscoveryApiError',
           'ApiAuthError', 'ResponseError', 'EmptyApiKeyError',
           'UnparsableApiResponseError']

from .error import ParameterError, HttpApiError, \
    DomainDiscoveryApiError, ApiAuthError, ResponseError, \
    EmptyApiKeyError, UnparsableApiResponseError



import datetime
from json import loads, JSONDecodeError
import re

from .net.http import ApiRequester
from .models.response import Response
from .exceptions.error import ParameterError, EmptyApiKeyError, \
    UnparsableApiResponseError


class Client:
    __default_url = "https://domains-subdomains-discovery.whoisxmlapi.com" \
                    "/api/v1"
    _api_requester: ApiRequester or None
    _api_key: str

    _re_api_key = re.compile(r'^at_[a-z0-9]{29}$', re.IGNORECASE)

    _PARSABLE_FORMAT = 'json'

    JSON_FORMAT = 'json'
    XML_FORMAT = 'xml'

    __DATETIME_OR_NONE_MSG = 'Value should be None or an instance of ' \
                             'datetime.date'

    def __init__(self, api_key: str, **kwargs):
        """
        :param api_key: str: Your API key.
        :key base_url: str: (optional) API endpoint URL.
        :key timeout: float: (optional) API call timeout in seconds
        """

        self._api_key = ''

        self.api_key = api_key

        if 'base_url' not in kwargs:
            kwargs['base_url'] = Client.__default_url

        self.api_requester = ApiRequester(**kwargs)

    @property
    def api_key(self) -> str:
        return self._api_key

    @api_key.setter
    def api_key(self, value: str):
        self._api_key = Client._validate_api_key(value)

    @property
    def api_requester(self) -> ApiRequester or None:
        return self._api_requester

    @api_requester.setter
    def api_requester(self, value: ApiRequester):
        self._api_requester = value

    @property
    def base_url(self) -> str:
        return self._api_requester.base_url

    @base_url.setter
    def base_url(self, value: str or None):
        if value is None:
            self._api_requester.base_url = Client.__default_url
        else:
            self._api_requester.base_url = value

    @property
    def timeout(self) -> float:
        return self._api_requester.timeout

    @timeout.setter
    def timeout(self, value: float):
        self._api_requester.timeout = value

    def get(self, **kwargs) -> Response:
        """
        Get parsed API response as a `Response` instance.

        :key domains: Required if domains aren't specified.
                Dictionary. Take a look at API documentation for the format
        :key subdomains: Required if domains aren't specified
                Dictionary. Take a look at API documentation for the format
        :key since_date: Optional. datetime.date. Min date by default.
        :return: `Response` instance
        :raises ConnectionError:
        :raises DomainDiscoveryApiError: Base class for all errors below
        :raises ResponseError: response contains an error message
        :raises ApiAuthError: Server returned 401, 402 or 403 HTTP code
        :raises BadRequestError: Server returned 400 or 422 HTTP code
        :raises HttpApiError: HTTP code >= 300 and not equal to above codes
        :raises ParameterError: invalid parameter's value
        """

        kwargs['output_format'] = Client._PARSABLE_FORMAT

        response = self.get_raw(**kwargs)

        try:
            parsed = loads(str(response))
            if 'domainsCount' in parsed:
                return Response(parsed)
            raise UnparsableApiResponseError(
                "Could not find the correct root element.", None)
        except JSONDecodeError as error:
            raise UnparsableApiResponseError(
                    "Could not parse API response",
                    error)

    def get_raw(self, **kwargs) -> str:
        """
        Get raw API response.

        :key domains: Required if subdomains aren't specified.
                Dictionary. Take a look at API documentation for the format
        :key subdomains: Required if domains aren't specified
                Dictionary. Take a look at API documentation for the format
        :key since_date: Optional. datetime.date. Min date by default.
        :key output_format: Optional. Use constants JSON_FORMAT and XML_FORMAT
        :return: str
        :raises ConnectionError:
        :raises DomainDiscoveryApiError: Base class for all errors below
        :raises ResponseError: response contains an error message
        :raises ApiAuthError: Server returned 401, 402 or 403 HTTP code
        :raises BadRequestError: Server returned 400 or 422 HTTP code
        :raises HttpApiError: HTTP code >= 300 and not equal to above codes
        :raises ParameterError: invalid parameter's value
        """

        if self.api_key == '':
            raise EmptyApiKeyError('')

        if 'domains' in kwargs:
            domains = Client._validate_domains(kwargs['domains'])
        else:
            domains = None

        if 'subdomains' in kwargs:
            subdomains = Client._validate_subdomains(kwargs['subdomains'])
        else:
            subdomains = None

        if not domains and not subdomains:
            raise ParameterError(
                "Required either domains or subdomains")

        if 'response_format' in kwargs:
            kwargs['output_format'] = kwargs['response_format']
        if 'output_format' in kwargs:
            output_format = Client._validate_output_format(
                kwargs['output_format'])
        else:
            output_format = Client._PARSABLE_FORMAT

        if 'since_date' in kwargs:
            since_date = Client._validate_date(kwargs['since_date'])
        else:
            since_date = Client._validate_date(datetime.date.min)

        return self._api_requester.post(self._build_payload(
            self.api_key,
            domains,
            subdomains,
            since_date,
            output_format
        ))

    @staticmethod
    def _build_payload(
            api_key,
            domains,
            subdomains,
            since_date,
            output_format,
    ) -> dict:
        tmp = {
            'apiKey': api_key,
            'domains': domains,
            'subdomains': subdomains,
            'sinceDate': since_date,
            'outputFormat': output_format,
        }

        payload = {}
        for k, v in tmp.items():
            if v is not None:
                payload[k] = v
        return payload

    @staticmethod
    def _validate_api_key(api_key) -> str:
        if Client._re_api_key.search(
                str(api_key)
        ) is not None:
            return str(api_key)
        else:
            raise ParameterError("Invalid API key format.")

    @staticmethod
    def _validate_date(value: datetime.date or None):
        if value is None or isinstance(value, datetime.date):
            return str(value)

        raise ParameterError(Client.__DATETIME_OR_NONE_MSG)

    @staticmethod
    def _validate_domains(value) -> dict:
        if value is None:
            raise ParameterError("Domain list cannot be None.")
        else:
            return Client._validate_terms(value)

    @staticmethod
    def _validate_output_format(value: str):
        if value.lower() in [Client.JSON_FORMAT, Client.XML_FORMAT]:
            return value.lower()

        raise ParameterError(
            f"Response format must be {Client.JSON_FORMAT} "
            f"or {Client.XML_FORMAT}")

    @staticmethod
    def _validate_subdomains(value) -> dict:
        if value is None:
            raise ParameterError("Subdomain list cannot be None.")
        else:
            return Client._validate_terms(value)

    @staticmethod
    def _validate_terms(value) -> dict:
        include, exclude = [], []
        if type(value) is dict:
            if 'include' in value:
                include = list(map(lambda s: str(s), value['include']))
                include = list(
                    filter(lambda s: s is not None and len(s) > 0, include))
                if 4 <= len(include) <= 1:
                    raise ParameterError("Include terms list must have "
                                         "from 1 to 4 terms.")
            if 'exclude' in value:
                exclude = list(map(lambda s: str(s), value['exclude']))
                exclude = list(
                    filter(lambda s: s is not None and len(s) > 0, exclude))
                if 4 <= len(exclude) <= 0:
                    raise ParameterError("Exclude terms list must have "
                                         "from 0 to 4 terms.")
            if include:
                return {'include': include, 'exclude': exclude}

        raise ParameterError("Expected a dict with 2 lists of strings.")



import datetime
import os
import unittest
from domaindiscovery import Client
from domaindiscovery import ParameterError, ApiAuthError


class TestClient(unittest.TestCase):
    """
    Final integration tests without mocks.

    Active API_KEY is required.
    """

    def setUp(self) -> None:
        self.client = Client(os.getenv('API_KEY'))
        self.correct_domains = {
            'include': ['example.com']
        }

        self.incorrect_domains = {
            'include': []
        }

    def test_get_correct_data(self):
        response = self.client.get(domains=self.correct_domains)
        self.assertIsNotNone(response.domains_count)

    def test_extra_parameters(self):
        response = self.client.get(
            domains=self.correct_domains,
            since_date=datetime.date.today() - datetime.timedelta(days=10)
        )
        self.assertIsNotNone(response.domains_count)

    def test_empty_terms(self):
        with self.assertRaises(ParameterError):
            self.client.get()

    def test_empty_api_key(self):
        with self.assertRaises(ParameterError):
            client = Client('')
            client.get(domains=self.correct_domains)

    def test_incorrect_api_key(self):
        client = Client('at_00000000000000000000000000000')
        with self.assertRaises(ApiAuthError):
            client.get(domains=self.correct_domains)

    def test_raw_data(self):
        response = self.client.get_raw(
            domains=self.correct_domains,
            output_format=Client.XML_FORMAT)
        self.assertTrue(response.startswith('<?xml'))

    def test_subdomains(self):
        response = self.client.get(
            subdomains=self.correct_domains
        )
        self.assertIsNotNone(response.domains_count)

    def test_incorrect_domains(self):
        with self.assertRaises(ParameterError):
            self.client.get(domains=self.incorrect_domains)

    def test_incorrect_subdomains(self):
        with self.assertRaises(ParameterError):
            self.client.get(subdomains=self.incorrect_domains)

    def test_incorrect_date(self):
        with self.assertRaises(ParameterError):
            self.client.get(domains=self.correct_domains,
                            since_date='19-19-19')

    def test_output(self):
        with self.assertRaises(ParameterError):
            self.client.get(domains=self.correct_domains,
                            response_format='yaml')


if __name__ == '__main__':
    unittest.main()
